/* 
 * Copyright (C) 2008 Laird Breyer
 *  
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Author:   Laird Breyer <laird@lbreyer.com>
 */

#include "boxreader.h"
#include "util.h"
#include "bittwiddler.h"
#include "tessboxes.h"

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

extern options_t u_options;

bool_t create_boxreader(boxreader_t *p, int fd) {
  if( p ) {
    
    clear_box(&p->cur);
    /* duplicate the file descriptor, because gzclose will destroy it */
    p->f = gzdopen(dup(fd), "rb");
    if( p->f ) {
      p->bufsize = 4096;
      /* buffer has one extra byte for a NULL marker */
      p->buf = malloc((p->bufsize + 1) * sizeof(byte_t));
      if( p->buf ) {
	p->buf[p->bufsize] = '\0'; /* for string functions */
	p->bufpos = p->buf; /* read position */
	p->bufend = p->buf; /* end of data */
      } else {
	p->bufsize = 0;
      }
      if( !get_header_info_boxreader(p) ) {
	free_boxreader(p);
	return FALSE;
      }
      return TRUE;
    }
  }
  return FALSE;
}

bool_t get_header_info_boxreader(boxreader_t *p) {
  /* no actual header, but a good excuse to call read_first_buffer */
  if( p &&
      read_first_buffer_boxreader(p) ) {
    return TRUE;
  }
  return FALSE;
}

bool_t free_boxreader(boxreader_t *p) {
  if( p ) {
    if( p->f ) {
      gzclose(p->f);
      p->f = NULL;
    }
    if( p->buf ) {
      free(p->buf);
      p->buf = NULL;
    }
  }
  return FALSE;
}

bool_t read_integer_boxreader(boxreader_t *p, unsigned int *retval) {
  char *q;
  if( p && retval && (p->bufend > p->bufpos) ) {
    /* this uses the NULL at the end of the buffer */
    *retval = (unsigned int)strtoul((const char *)p->bufpos, &q, 10); 
    if( ((byte_t *)q > p->bufpos) && ((byte_t *)q <= p->bufend) ) {
      p->bufpos = (byte_t *)q;
      return TRUE;
    }
  }
  return FALSE;
}

bool_t read_first_buffer_boxreader(boxreader_t *p) {
  if( p && p->f ) {
    gzseek(p->f, 0, SEEK_SET);
    return read_next_buffer_boxreader(p);
  } 
  return FALSE;
}

bool_t read_next_buffer_boxreader(boxreader_t *p) {
  int n;
  if( p && p->f && p->buf ) {
    n = gzread(p->f, p->buf, p->bufsize);
    if( n == -1 ) {
      errormsg(E_WARNING, "file read error\n");
      free_boxreader(p);
    } else {
      p->bufpos = p->buf;
      p->bufend = p->buf + n;
    }
    return (n>0);
  }
  return FALSE;
}

void read_record_boxreader(boxreader_t *p, const byte_t *line, annotations_t *a) {
  byte_t *q = p->cur.symbol;
  while( *line && !isspace(*line) && (q < (p->cur.symbol + SYMLEN)) ) {
    *q++ = *line++;
  }
  *q = '\0';
  while( *line && !isdigit(*line) ) { line++; }
  p->cur.left = strtol((char *)line, (char **)&q, 10);
  line = q;
  while( *line && !isdigit(*line) ) { line++; }
  p->cur.top = strtol((char *)line, (char **)&q, 10);
  line = q;
  while( *line && !isdigit(*line) ) { line++; }
  p->cur.right = strtol((char *)line, (char **)&q, 10);
  line = q;
  while( *line && !isdigit(*line) ) { line++; }
  p->cur.bottom = strtol((char *)line, (char **)&q, 10);
  if( u_options & FL_TESSV2 ) {
    p->cur.page = 0; /* legacy v2 file format */
  } else {
    line = q;
    while( *line && !isdigit(*line) ) { line++; }
    p->cur.page = strtol((char *)line, (char **)&q, 10);
  }
  while( *line && *line != '#' ) { line++; }
  p->cur.annotation = 
    ( (*line == '#' && isdigit(line[1])) ? (line[1] - '0') : 0 );
  if( *line == '#' && isdigit(line[1]) && isspace(line[2]) ) {
    set_annotations(a, p->cur.annotation, line + 3);
  }
}


bool_t parse_boxreader(boxreader_t *p, imagebuffer_t *ib, boxlist_t *bl, annotations_t *an) {
  /* unsigned int rowcounter; */
#define LINELEN 128
  byte_t line[LINELEN+1];
  byte_t *linepos;

  if( p && p->f && p->buf && (p->bufsize > 0) && ib ) {
    seekpos_imagebuffer(ib, 0, 0);
    /* rowcounter = 0; */
    line[LINELEN] = '\0'; 
    linepos = line;
    do {
      do {
	*linepos = *p->bufpos;
	if( (*p->bufpos == '\n') || (*p->bufpos == '\r') ||
	    (linepos > line + LINELEN) ) {
	  *linepos = '\0';
	  read_record_boxreader(p, line, an);
	  if( !append_boxlist(bl, &p->cur) ) {
	    return FALSE;
	  }
	  if( !box_bw_imagebuffer(ib, 
				  p->cur.left, 
				  ib->height - p->cur.bottom, 
				  p->cur.right, 
				  ib->height - p->cur.top, 
				  1, 1) ) {
	    return FALSE;
	  }
	  linepos = line;
	} else {
	  linepos++;
	}
	p->bufpos++;

      } while( p->bufpos < p->bufend );
    } while( read_next_buffer_boxreader(p) );
    
    /* file was automatically closed on error */
    return (p->f != NULL);
  }
  return FALSE;
}
